
local AddOnName = "AutoProfitX";
local DBName = "AutoProfitXDB";
local DBSName = "AutoProfitXDBS";
local Version = GetAddOnMetadata( AddOnName, "Version" );

-- Create our addon.
AutoProfitX = LibStub("AceAddon-3.0"):NewAddon(AddOnName, 
                                              "AceConsole-3.0", 
                                              "AceEvent-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale(AddOnName, false);

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)

local tooltip = LibStub("LibGratuity-3.0", true)

_G["apxAuthor"] = {
     ["Lindarena@Laughing Skull"] = true,
     ["Nagdand@Laughing Skull"] = true,
}

AutoProfitXButtonSpinner = {
     { name = L["Never spin"], value = 0, },
     { name = L["Mouse-over and profit"], value = 1, },
     { name = L["Mouse-over"], value = 2, },
     { name = L["Profits"], value = 3, },
}

local apxf = CreateFrame( "Frame", "AutoProfitXFrame", UIParent )
local name, realm
local totalProfit = 0
local IES = "Equipment Sets: %s"
--make proficiency table local (defined in proficiencies.lua)
local prof = AutoProfitX_Proficiencies[UnitClass("player")] or {}
AutoProfitX_Proficiencies = nil

--some strings commonly used in addon
local linkMatch = "|c%x+|Hitem:[%-%d:]*|h%[.-%]|h|r"
local classMatch = string.format( ITEM_CLASSES_ALLOWED, "([%w, ]*)" )
local equipMatch = string.format( IES, "([%w, ]*)" )

local sCmds = {
     "autoprofitx",
     "autoprofit",
     "apx",
}

local sOpts = {
     name = "AutoProfitX Command Handler",
     type = "group",
     handler = AutoProfitX,
     args = {
          addl = {
               type = "execute", order = 1,
               name = L["addl"],
               desc = L["Add Local exception. Usage : <item link>[<item link>...]"],
               func = function( self )
                    local _, arg1 = strmatch( string.gsub( self.input, "addl ", ":" ), "(.*):%s(.*)" )
                    AutoProfitX:idUtils( "idAdd", "local", arg1 );
               end,
          },
          addg = {
               type = "execute", order = 2,
               name = L["addg"],
               desc = L["Add Global exception. Usage : <item link>[<item link>...]"],
               func = function( self )
                    local _, arg1 = strmatch( string.gsub( self.input, "addg ", ":" ), "(.*):%s(.*)" )
                    AutoProfitX:idUtils( "idAdd", "global", arg1 );
               end,
          },
          reml = {
               type = "execute", order = 3,
               name = L["reml"],
               desc = L["Remove Local exception. Usage : <item link>[<item link>...]"],
               func = function( self )
                    local _, arg1 = strmatch( string.gsub( self.input, "reml ", ":" ), "(.*):%s(.*)" )
                    AutoProfitX:idUtils( "idRemove", "local", arg1 );
               end,
          },
          remg = {
               type = "execute", order = 4,
               name = L["remg"],
               desc = L["Remove Global exception. Usage : <item link>[<item link>...]"],
               func = function( self )
                    local _, arg1 = strmatch( string.gsub( self.input, "remg ", ":" ), "(.*):%s(.*)" )
                    AutoProfitX:idUtils( "idRemove", "global", arg1 );
               end,
          },
          geList = {
               type = "execute", order = 5,
               name = L["list"],
               desc = L["Lists Your Global exceptions List."],
               func = function()
                    AutoProfitX:ListExceptions( "Global" );
               end
          },
          leList = {
               type = "execute", order = 6,
               name = L["list"],
               desc = L["Lists Your Local exceptions List."],
               func = function()
                    AutoProfitX:ListExceptions( "Local" );
               end
          },

          options = {
               type = "execute", order = 7,
               name = L["options"],
               desc = L["Show options panel."],
               func = function()
                  InterfaceOptionsFrame_OpenToCategory(AutoProfitX.ConfigFrames.Options);
               end,
          },
     },
}

local apxAbout = {
     type = "group",
     name = "About",
     args = {
          confgendesc = {
               type = "description",
               name = L["Automatically sell junk items when you click the button on the vendors window."],
               cmdHidden = true,
	       order = 1,
          },
          confinfodesc = {
               name = L["About"],
               type = "group", order = 2, inline = true,
               args = {
                    confversiondesc = {
                         order = 1,
                         type = "description",			
                         name = "|cffffd700Version : |cff20ff20"..Version,
                         cmdHidden = true,
                    },
                    confauthordesc = {
                         order = 2,
                         type = "description",
                         name = "|cffffd700Author : |cffff8c00"..GetAddOnMetadata("AutoProfitX", "Author"),
                         cmdHidden = true,
                    },
                    confcatdesc = {
                         order = 4,
                         type = "description",
                         name = "|cffffd700Category : |cffffffff"..GetAddOnMetadata("AutoProfitX", "X-Category"),
                         cmdHidden = true,
                    },
                    confemaildesc = {
                         order = 5,
                         type = "description",
                         name = "|cffffd700Email : |cffffffff"..GetAddOnMetadata("AutoProfitX", "X-Email"),
                         cmdHidden = true,
                    },
                    confwebsitedesc = {
                         order = 6,
                         type = "description",
                         name = "|cffffd700Website : |cffffffff"..GetAddOnMetadata("AutoProfitX", "X-Website"),
                         cmdHidden = true,
                    },
               },
          },
	  loadmsg = {
               type = "toggle", order = 3,
	       name = L["Addon Load Message"],
	       desc = L["This will display a message in the chat frame when then addon is fully loaded."],
	       get = function() return AutoProfitX.db.global.showGreeting; end,
	       set = function( _, value )
	            if ( value ) then
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Addon Load Message"], "on" ) );
		    else
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Addon Load Message"], "off" ) );
		    end
	            AutoProfitX.db.global.showGreeting = value; 
               end,
	  },
          hidemini = {
               type = "toggle",
               name = L["Hide Minimap Button"],
               desc = L["This will remove the button from the minimap."],
               get = function() return AutoProfitX.db.global.MinimapIcon.hide; end,
               set = function( _, value )
                    if ( not value ) then
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Minimap Icon"], "on" ) );
                         LDBIcon:Show( AddOnName )
                    else
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Minimap Icon"], "off" ) );
                         LDBIcon:Hide( AddOnName )
                    end
                    AutoProfitX.db.global.MinimapIcon.hide = value; 
               end,
               order = 4,
          },
	  memusage = {
               type = "execute",
	       name = L["Memory Usage"],
	       desc = L["This will display memory usgae in the chat frame."],
	       func = function()
	            UpdateAddOnMemoryUsage();
		    AutoProfitX:Print( "print", L["is using|cff304aa9 %s kb|r of memory."]:format( floor(GetAddOnMemoryUsage(AddOnName)) ) )
	       end,
               order = 5,
	  },
     },
}

local apxOpt = {
     type = "group",
     name = L["Options"],
     order = 1,
     args = {
          autoSell = {
               name = L["Auto Sell"],
               type = "toggle", order = 1,
               desc = L["Automatically sell junk items when opening vendor window."],
               get = function () return AutoProfitX.dbpc.char.autoSell;  end,
               set = function( _, value )
                    if ( value ) then
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Auto Sell"], "on" ) );
                    else
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Auto Sell"], "off" ) );
                    end
                    AutoProfitX.dbpc.char.autoSell = value;
               end,
          },
          sellSoulbound = {
               name = L["Sell Soulbound"],
               type = "toggle", order = 2 ,
               desc = L["Sell unusable soulbound items."],
               get = function () return AutoProfitX.dbpc.char.checkSoulbound; end,
	       set = function ( _, value )
                    if ( value ) then
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Sell Soulbound"], "on" ) );
                    else
                         AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Sell Soulbound"], "off" ) );
                    end
                    AutoProfitX.dbpc.char.checkSoulbound = value;
               end,
          },
          openBags = {
               name = L["Open Bags"],
               type = "toggle", order = 3,
               desc = L["This will open all bags when you visit a vendor."],
               get = function () return AutoProfitX.dbpc.char.openBags; end,
	       set = function ( _, value )
                    local bagCheck = AutoProfitX:otherBag()
                    if ( not bagCheck ) then
                         if ( value ) then
                              AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Open Bags"], "on" ) );
                         else
                              AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Open Bags"], "off" ) );
                         end
                         AutoProfitX.dbpc.char.openBags = value;
                    else
                         AutoProfitX:Print( "print", L["Found a bag mod please use it's auto open function. Disableing"] )
                    end
               end,
          },
          reportSales = {
               type = "group", order = 4, inline = true,
               name = L["Sales Reports"],
               args = {
                    silent = {
                         name = L["Sales Reports"],
                         type = "toggle", order = 1,
                         desc = L["Print items being sold in chat frame."],
                         get = function () return AutoProfitX.dbpc.char.silent; end,
                         set = function ( _, value )
                              if ( value ) then
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Sales Reports"], "on" ) );
                              else
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Sales Reports"], "off" ) );
                              end
                              AutoProfitX.dbpc.char.silent = value;
                         end,
                    },
                    fancysales = {
                         name = L["Fancy Sales Reports"],
                         type = "toggle", order = 2,
                         desc = L["Print items being sold in chat frame in coins."],
                         get = function () return AutoProfitX.dbpc.char.fancysales; end,
                         set = function ( _, value )
                              if ( value ) then
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Auto Sell"], "on" ) );
                              else
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Auto Sell"], "off" ) );
                              end
                              AutoProfitX.dbpc.char.fancysales = value; end,
                    },
                    stackcount = {
                         name = L["Include Stack Count"],
                         type = "toggle", order = 3,
                         desc = L["This will show the stack count int he sales report."],
                         get = function () return AutoProfitX.dbpc.char.stackCount; end,
                         set = function ( _, value )
                              if ( value ) then
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Include Stack Count"], "on" ) );
                              else
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Include Stack Count"], "off" ) );
                              end
                              AutoProfitX.dbpc.char.stackCount = value;
                         end,
                    },
               },
          },
          profits = {
               type = "group", order = 5, inline = true,
               name = L["Show Profits"],
	       args = {
                    profit = {
                         name = L["Show Profit"],
                         type = "toggle", order = 1,
                         desc = L["Print total profit after sale."],
                         get = function () return AutoProfitX.dbpc.char.showTotal; end,
                         set = function ( _, value )
                              if ( value ) then
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Show Profits"], "on" ) );
                              else
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Show Profits"], "off" ) );
                              end
                              AutoProfitX.dbpc.char.showTotal = value;
                         end,
                    },
                    fancyprofits = {
                         name = L["Show Fancy Profits"],
                         type = "toggle", order = 2,
                         desc = L["Print total profit after sale using coins."],
                         get = function () return AutoProfitX.dbpc.char.fancyprofits; end,
                         set = function ( _, value )
                              if ( value ) then
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Show Fancy Profits"], "on" ) );
                              else
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Show Fancy Profits"], "off" ) );
                              end
                              AutoProfitX.dbpc.char.fancyprofits = value;
                         end,
                    },
                    buttonSpin = {
                         name = L["Button Animation Options"],
                         type = "select", order = 3,
                         desc = L["Set up when you want the treasure pile in the button to spin."],
                         get = function() return AutoProfitX.dbpc.char.buttonSpin; end,
                         set = function( _, value ) AutoProfitX.dbpc.char.buttonSpin = value; end,
                         values = function()
                              local preList = {};
                              local v;
                              for _, v in pairs( AutoProfitXButtonSpinner ) do
                                   if ( v.value ~= AutoProfitX.dbpc.char.buttonSpin ) then
                                        preList[v.value] = "|cffffff9a"..v.name.."|r"
                                   else
                                        preList[v.value] = "|cff19ff19"..v.name.."|r"
                                   end
                              end
                              return preList;
                         end
                    },
                    buttonSpinRate = {
                         name = L["Button Gold Pile Spin Rate"],
                         type = "range", order = 4,
                         desc = L["This will change the speed at witch the pile of gold spins if you have stuff to sell."],
                         min = 0,
                         max = 1,
                         step = .1,
                         get = function() return AutoProfitX.dbpc.char.spinRate; end,
                         set = function( _, value )
                              AutoProfitX_SellButton_TreasureModel.rotRate = value;
                              AutoProfitX.dbpc.char.spinRate = value;
                         end,
                    },
               },
          },
          dockButton = {
               name = L["Reset Button Pos"],
               type = "execute", order = 6,
               desc = L["Reset APX button position on the vendor screen to the top right corner."],
               func = function ()
                    AutoProfitX:Print( "print", L["Resetting Auto Sell Button."] )
                    AutoProfitX:SetButtonXY( -41, -37 ) 
               end,
          },
     },
}

local apxExc = {
     type = "group",
     name = L["Exception List"],
     args = {
	  execpt = {
	       type = "group",
	       name = L["Vendor Exceptions"],
	       inline = true,
	       order = 7,
	       args = {
		    header1 = {
                         type = "header",
			 name = L["Clear Exceptions"],
			 order = 1,
		    },
	            btn1 = {
                         type = "execute",
	                 name = L["Global Exceptions"],
                         desc = L["This will clear anything that you have added to the global exceptions for AutoSell"],
	                 func = function()
                              wipe(AutoProfitX.db.global.geList)
                              AutoProfitX:Print( "print", format( L["|cff304aa9%s|r exceptions have been reset."], "Global" ) );
			 end,
	                 order = 2,
	            },
	            btn2 = {
	                 type = "execute",
	                 name = L["Character Exceptions"],
	                 desc = L["This will clear anything that you have added the the character exceptions for AutoSell"],
	                 func = function()
                              wipe(AutoProfitX.dbpc.char.leList)
                              AutoProfitX:Print( "print", format( L["|cff304aa9%s|r exceptions have been reset."], "Local" ) );
			 end,
	                 order = 3,
	            },
		    header2 = {
                         type = "header",
			 name = L["Global Exceptions"],
			 order = 4,
		    },
		    desc1 = {
                         type = "description",
			 name = L["Drag and drop the item into the feild of the action you wish to perform."],
			 order = 5,
		    },
		    addglobal = {
                         type = "input",
			 name = L["Add Global Item : "],
			 desc = L["This will add a item to the global exceptions list"],
			 usage = L["|cff19ff19<Item Link>|r"],
			 get = false,
			 set = function( _, value ) AutoProfitX:idUtils( "idAdd", "global", value ); end,
			 order = 6,
		    },
		    removeglobal = {
                         type = "input",
			 name = L["Remove Global Item : "],
			 desc = L["This will remove a item from the global exceptions list"],
			 usage = L["|cff19ff19<Item Link>|r"],
			 get = false,
			 set = function( _, value ) AutoProfitX:idUtils( "idRemove", "global", value ) end,
			 order = 7,
		    },
		    header3 = {
                         type = "header",
			 name = L["Character Exceptions"],
			 order = 8,
		    },
		    desc2 = {
                         type = "description",
			 name = L["Drag and drop the item into the feild of the action you wish to perform."],
			 order = 9,
		    },
		    addcharacter = {
                         type = "input",
			 name = L["Add Character Item : "],
			 desc = L["This will add a item from the character exceptions list for this character"],
			 usage = L["|cff19ff19<Item Link>|r"],
			 get = false,
			 set = function( _, value ) AutoProfitX:idUtils( "idAdd", "local", value ); end,
			 order = 10,
		    },
		    removecharacter = {
                         type = "input",
			 name = L["Remove Character Item : "],
			 desc = L["This will remove a item from the character exceptions list for this character"],
			 usage = L["|cff19ff19<Item Link>|r"],
			 get = false,
			 set = function( _, value ) AutoProfitX:idUtils( "idRemove", "local", value ); end,
			 order = 11,
		    },
	       },
	  },
     },
}

local apxDebuging = {
     name = L["AutoProfitX Debuging"],
     type = "group",
     args = {
          confinfodesc = {
               name = L["Debug Setting"],
               type = "group", inline = true,
               order = 3,
               args = {
                    confgendesc = {
                         type = "description",
                         name = L["You should only use this when you are having errors to see if this addon is causing the problem."],
                         cmdHidden = true,
                         order = 1,
                    },
                    optdebug = {
                         type = "toggle",
                         width = "full",
                         name = L["Debug"],
                         desc = L["This will turn on internal debugging"],
                         get = function() return AutoProfitX.db.global.debug.state; end,
                         set = function( _, value )
                              if ( value ) then
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Debugging"], "on" ) );
                              else
                                   AutoProfitX:Print( "print", L["%s now|cff304aa9 %s|r."]:format( L["Debugging"], "off" ) );
                              end
                              AutoProfitX.db.global.debug.state = value; 
                         end,
                         order = 4,
                    },
               },
          },
     },
}

local Defaults = {
     global = {
          ["geList"] = {
          },
          ["showGreeting"] = true,
          ["MinimapIcon"] = {
               ["hide"] = false,
          },
          ["debug"] = {
               ["state"] = false,
          },
          ["otherBag"] = false,
          ["bag"] = {
               [0] = false,
               [1] = false,
               [2] = false,
               [3] = false,
               [4] = false,
          },
     },
}

local DefaultsPC = {
     char = {
          ["autoSell"] = false,
          ["silent"] = false,
          ["fancysals"] = false,
          ["showTotal"] = true,
          ["buttonYpos"] = -37,
          ["buttonXpos"] = -41,
          ["checkSoulbound"] = false,
          ["buttonSpin"] = 3,
          ["spinRate"] = 0.6,
          ["fancyprofits"] = false,
          ["stackCount"] = false,
          ["openBags"] = false,
          ["leList"] = {
          },
     },

}

-- AutoProfitX:OnInitialize()
-- =========================================================
function AutoProfitX:OnInitialize()
     -- Set up our database
     self.db = LibStub("AceDB-3.0"):New(DBName, Defaults, "Default");
     self.dbpc = LibStub("AceDB-3.0"):New(DBSName, DefaultsPC, "Default");

     -- Set up our config options.
     local profiles      = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db);
     local AutoProfitXConf = LibStub("AceConfig-3.0");
     local AutoProfitXReg  = LibStub("AceConfigRegistry-3.0");
     local AutoProfitXDiag = LibStub("AceConfigDialog-3.0");

     name = GetUnitName("player")
     realm = GetCVar("realmName")
     local curUser = name.."@"..realm;

     AutoProfitXConf:RegisterOptionsTable( AddOnName, sOpts, sCmds );

     AutoProfitXReg:RegisterOptionsTable( AddOnName.." About", apxAbout );
     AutoProfitXReg:RegisterOptionsTable( AddOnName.." Options", apxOpt );
     AutoProfitXReg:RegisterOptionsTable( AddOnName.." Exceptions", apxExc );
     if ( _G["apxAuthor"][curUser] ) then AutoProfitXReg:RegisterOptionsTable( AddOnName.." Debuging", apxDebuging); end

     self.ConfigFrames = {
          About = AutoProfitXDiag:AddToBlizOptions( AddOnName.." About", AddOnName ),
          Options = AutoProfitXDiag:AddToBlizOptions( AddOnName.." Options", "Options", AddOnName ),
          Exceptions = AutoProfitXDiag:AddToBlizOptions( AddOnName.." Exceptions", "Exceptions", AddOnName ),
     }

     if ( _G["apxAuthor"][curUser] ) then self.ConfigFrames.Debuging = AutoProfitXDiag:AddToBlizOptions( AddOnName.." Debuging", "Debuging", AddOnName ); end

     self:RegisterEvent("PLAYER_LOGIN");
     self:RegisterEvent("PLAYER_ENTERING_WORLD");

     self.apxf = apxf;
end

-- AutoProfitX:OnEnable()
-- =========================================================
function AutoProfitX:OnEnable()
     self:RegisterEvent("MERCHANT_SHOW")
end

-- AutoProfitX:OnDisable()
-- =========================================================
function AutoProfitX:OnDisable()
     self:UnregisterEvent("MERCHANT_SHOW");
     self:UnregisterEvent("BAG_UPDATE");
     self:RegisterEvent("PLAYER_LOGIN");
     self:RegisterEvent("PLAYER_ENTERING_WORLD");
end

-- AutoProfitX:PLAYER_ENTERING_WORLD()
-- =========================================================
function AutoProfitX:PLAYER_ENTERING_WORLD()
     local bagCheck = AutoProfitX:otherBag()
     if ( bagCheck ) then AutoProfitX.dbpc.char.openBags = false; end
     AutoProfitX_SellButton:SetPoint( "TOPRIGHT", "MerchantFrame", "TOPRIGHT", AutoProfitX.dbpc.char.buttonXpos, AutoProfitX.dbpc.char.buttonYpos );
end

-- AutoProfitX:MERCHANT_SHOW()
-- =========================================================
function AutoProfitX:MERCHANT_SHOW()
     if ( AutoProfitX.dbpc.char.openBags ) then
          AutoProfitX:Print("Opening All Bags")
          AutoProfitX:OpenAllBags( false )
     end
     if ( AutoProfitX.dbpc.char.autoSell ) then
          self:Print( "bug", "autoSell is on" )
          local profit = AutoProfitX:GetProfit()
          if ( profit > 0 ) then
               AutoProfitX:SellJunk()
               if ( AutoProfitX.dbpc.char.showTotal ) then
                    if ( AutoProfitX.dbpc.char.fancyprofits ) then 
                         self:Print( "print", format( L["Total profits: %s"], AutoProfitX:GetMoneyString( profit ) ) );
                    else
                         self:Print( "print", format( L["Total profits: %s"], AutoProfitX:CopperToGold( profit ) ) );
                    end
               end
          end
     else
          --register BAG_UPDATE event for updating button
          self.apxf:RegisterEvent("BAG_UPDATE")
          self.apxf:RegisterEvent("MERCHANT_CLOSED")
     end
end

-- AutoProfitX:MERCHANT_CLOSED()
-- =========================================================
function AutoProfitX:MERCHANT_CLOSED()
     if ( AutoProfitX.dbpc.char.openBags ) then
          AutoProfitX:Print("Closing All Bags")
          AutoProfitX:OpenAllBags( true )
     end
     -- unregister BAG_UPDATE event used for updating icon
     -- --------------------------------------------------
     if ( self.apxf:IsEventRegistered("BAG_UPDATE") ) then
          self.apxf:UnregisterEvent("MERCHANT_CLOSED")
          self.apxf:UnregisterEvent("BAG_UPDATE")
     end
end

-- AutoProfitX:BAG_UPDATE()
-- =========================================================
function AutoProfitX:BAG_UPDATE()
     if ( AutoProfitX_SellButton:IsVisible() ) then
          AutoProfitX:OnShowButton()
     end
end

-- AutoProfitX:PLAYER_LOGIN()
-- =========================================================
function AutoProfitX:PLAYER_LOGIN()
     -- LDB launcher
     if ( LDB ) then
          AutoProfitXLauncher = LDB:NewDataObject("AutoProfitX", {
               type = "launcher",
               icon = "Interface\\Icons\\INV_Misc_Bag_EnchantedMageweave",
               OnClick = function( clickedframe, button )
                    if ( button == "RightButton" ) then AutoProfitX:ButtonClick( "config" ) else AutoProfitX:ButtonClick( "list" ) end
               end,
               OnTooltipShow = function( tt )
                    tt:AddLine( L["|cff304aa9[AutoProfitX]|r -|cffffffff %s|r"]:format( Version ) );
                    tt:AddLine( L["|cff304aa9Click|r|cffffffff to see the AddOn Info.|r"] );
                    tt:AddLine( L["|cff304aa9Right-click|r|cffffffff to open the options panel|r"] );
               end,
          })
          if ( LDBIcon and not IsAddOnLoaded( "Broker2FuBar" ) and not IsAddOnLoaded( "FuBar") ) then
               LDBIcon:Register( "AutoProfitX", AutoProfitXLauncher, AutoProfitX.db.global.MinimapIcon );
          end
     end
end

-- AutoProfitX:idUtils( )
-- =========================================================
function AutoProfitX:idUtils( action, type, exceptions )
     local itemID, name
     if ( action == "idCheck" ) then
          itemID = AutoProfitX:GetID( link )
          if ( AutoProditX.db.global.geList[itemID] ) then return true; end
          if ( AutoProfitX.db.char.leList[itemID] ) then return true; end
          return false;
     elseif ( action == "idRemove" ) then
          for link in string.gmatch( exceptions, linkMatch ) do
               name = GetItemInfo( link );
	       itemID = AutoProfitX:GetID( link )
               if ( itemID ) then
                    if ( type == "global") then
                         AutoProfitX:Print( "print", format( L["Removed|cff304aa9 %s |rfrom|cff304aa9 %s |rexceptions list"], name, "Global" ) )
                         AutoProfitX.db.global.geList[itemID] = nil;
                    elseif( type == "local" ) then
                         AutoProfitX:Print( "print", format( L["Removed|cff304aa9 %s |rfrom|cff304aa9 %s |rexceptions list"], name, "Character" ) )
                         AutoProfitX.dbpc.char.leList[itemID] = nil;
                    end
               else
                    AutoProfitX:Print( "print", L["Invalid item link provided."] )
               end
          end
     elseif ( action == "idAdd" ) then
          for link in string.gmatch( exceptions, linkMatch ) do
               name = GetItemInfo( link );
	       itemID = AutoProfitX:GetID( link )
               if ( itemID ) then
                    if ( type == "global") then
	                 AutoProfitX:Print( "print", format( L["Added|cff304aa9 %s |rto|cff304aa9 %s |rexceptions list"], name, "Global" ) )
                         AutoProfitX.db.global.geList[itemID] = link;
                    elseif( type == "local" ) then
                         AutoProfitX:Print( "print", format( L["Added|cff304aa9 %s |rto|cff304aa9 %s |rexceptions list"],  name, "Character" ) )
                         AutoProfitX.dbpc.char.leList[itemID] = link;
                    end
               else
                    AutoProfitX:Print( "print", L["Invalid item link provided."] )
               end
          end
     end
     if ( AutoProfitX_SellButton:IsVisible() ) then
          AutoProfitX:OnShowButton()
     end
end

-- AutoProfitX:ListExceptions()
-- =========================================================
function AutoProfitX:ListExceptions( lType )
     local eList
     if ( lType == "Global" ) then
          eList = AutoProfitX.db.global.geList;
     elseif ( lType == "Local" ) then
          eList = AutoProfitX.dbpc.char.leList;
     end
     local link
     local dispHeader = true
     for i in pairs( eList ) do
          if ( dispHeader ) then
               self:Print( "print", L[lType.." Exceptions List :"] );
               dispHeader = false
          end
          _, link = GetItemInfo( i )
          if ( link ) then
               self:Print( "print", link );
          end
     end
     if ( dispHeader ) then
          self:Print( "print", L["Your "..lType.." exception list is empty."] );
     end
end

--[[ AutoProfitX:ImportExceptionList()
-- =========================================================
function AutoProfitX:ImportExceptionList( fromChar )
     local iName, iRealm = strsplit("@",fromChar)
     if ( iName ) then
          if ( self.dbpc.char[iRealm] and self.dbpc.char[iRealm][iName] and self.dbpc.char[iRealm][iName].exceptionList ) then
               eList = tdeepcopy(self.dbpc.char[iRealm][iName].exceptionList)
               self.dbpc.char[realm][name].exceptionList = eList
               self:Print( "print", format( L["Exception list imported from %s on %s."], iName, iRealm ) )
          else
               self:Print( "print", format( L["Exception list could not be found for %s on %s."], iName, iRealm ) )
          end
          return
     end
end --]]

-- AutoProfitX:GetID( link )
-- =========================================================
function AutoProfitX:GetID( link )
     return string.match( link, "item:(%d+)" )
end

-- AutoProfitX:SellJunk()
-- =========================================================
function AutoProfitX:SellJunk()
     local link, money
     if ( MerchantFrame:IsVisible() and MerchantFrame.selectedTab == 1 ) then
          for bag = 0, 4 do
               for slot = 1, GetContainerNumSlots(bag) do
                    --if slot not empty and item is junk
                    link = GetContainerItemLink(bag, slot)
                    if ( link and self:IsJunk( bag, slot, link ) ) then
                         --sell item
                         UseContainerItem( bag, slot )
                         --if sale reporting is turned on (not silent), display sale message
                         if ( AutoProfitX.dbpc.char.silent ) then
                              money = select( 11, GetItemInfo( link ) ) * select( 2, GetContainerItemInfo( bag, slot ) )
                              if ( AutoProfitX.dbpc.char.fancysales ) then
                                   if ( AutoProfitX.dbpc.char.stackCount ) then
                                        self:Print( "print",  format( L["Sold %s x%s for %s"],  link, GetItemCount( link ), self:GetMoneyString( money ) ) )
                                   else
                                        self:Print( "print", format( L["Sold %s for %s"], link, self:GetMoneyString( money ) ) )
                                   end
                              else
                                   if ( AutoProfitX.dbpc.char.stackCount ) then
                                        self:Print( "print", format( L["Sold %s x%s."], link, GetItemCount( link ) ) )                                   
                                   else
                                        self:Print( "print", format( L["Sold %s."], link ) )
                                   end
                              end
                         end
                    end
               end
          end
     end
end

-- AutoProfitX:IsJunk()
-- =========================================================
function AutoProfitX:IsJunk( bag, slot, link )
     local _, _, quality = GetItemInfo(link)
     local id = AutoProfitX:GetID(link)
     if ( AutoProfitX.dbpc.char.leList[id] or AutoProfitX.db.global.geList[id] ) then
          return quality ~= 0
     elseif ( quality == 0 ) then
          return true;
     elseif ( AutoProfitX.dbpc.char.checkSoulbound and bag and slot ) then
          if ( not self:IsUsable( bag, slot, link ) ) then
               return true;
          end
     end
     return false;
end

-- AutoProfitX:IsUsable()
-- =========================================================
function AutoProfitX:IsUsable( bag, slot, link )
     tooltip:SetBagItem( bag, slot );
     if ( not tooltip:Find( ITEM_SOULBOUND ) ) then
          return true;     --if it's not soulbound then you can always use it
     end
     if ( tooltip:Find( equipMatch ) ) then
          return true;          -- SLAM ON DA BREAKS -- this is part of someone gear set
     end
     local class = UnitClass("player")     --check if item has class requirement
     local _, _, classes = tooltip:Find( classMatch )
     if ( classes ) then
          local found = false
          classes = { strsplit( L["LIST_SEPARATOR"], classes ) }
          for _, c in ipairs( classes ) do
               if ( class == c ) then
                    found = true
               end
          end
          if ( not found ) then return false; end     --class not in the list so you can't use it
     end
     local _, _, _, _, _, iType, iSubType, _, iEquipLoc = GetItemInfo(link)     --check if class can ever use this item
     local tbl = prof[iType]
     if ( tbl and ( tbl[iSubType] or ( tbl.noOffhand and iEquipLoc == "INVTYPE_WEAPONOFFHAND" ) ) ) then
          return false;
     end
     --seems usable
     return true;
end

-- AutoProfitX:GetProfit()
-- =========================================================
function AutoProfitX:GetProfit()
     totalProfit = 0
     if ( MerchantFrame:IsVisible() ) then
          local bagSlots, link
          for bag = 0, 4 do
               bagSlots = GetContainerNumSlots( bag )
               if ( bagSlots > 0 ) then
                    for slot = 1, bagSlots do
                         link = GetContainerItemLink( bag, slot )
                         if ( link ) then
                              if ( AutoProfitX:IsJunk( bag, slot, link ) ) then
                                   AutoProfitX_Tooltip:SetBagItem( bag, slot )
                              end
                         end
                    end
               end
          end
     end
     return totalProfit
end

-- AutoProfitX:DragButton()
-- =========================================================
function AutoProfitX:DragButton()
     local MAX_RIGHT = -41
     local MAX_LEFT = -280
     local detatch = false
     local scale = MerchantFrame:GetEffectiveScale()
     local cursorOffset = 1
     local xpos, ypos = GetCursorPosition()
     xpos = xpos/scale
     ypos = ypos/scale
     local mwpos = MerchantFrame:GetRight()
     if ( IsShiftKeyDown() or AutoProfitX.dbpc.char.buttonXpos > MAX_RIGHT or AutoProfitX.dbpc.char.buttonXpos < MAX_LEFT or AutoProfitX.dbpc.char.buttonYpos ~= buttonY ) then
          detatch = true
     end
     xpos = xpos - mwpos - cursorOffset
     if ( detatch ) then
          local mwypos = MerchantFrame:GetTop()
          ypos = ypos - mwypos - cursorOffset
     else
          if ( xpos > MAX_RIGHT ) then
               xpos = MAX_RIGHT
          elseif ( xpos < MAX_LEFT ) then
               xpos = MAX_LEFT
          end
          ypos = nil
     end
     self:SetButtonXY( xpos, ypos )
end

-- AutoProfitX:OnEnterButton( self )
-- =========================================================
function AutoProfitX:OnEnterButton( self )
     GameTooltip:SetOwner( self, "ANCHOR_RIGHT" )
     GameTooltip:SetText( L["Sell Junk Items"] )
     local profit = AutoProfitX:GetProfit()
     if ( AutoProfitX.dbpc.char.buttonSpin == 2 ) then AutoProfitX:UpdateButton() end
     if ( profit > 0 ) then
          if ( AutoProfitX.dbpc.char.buttonSpin == 1 ) then AutoProfitX:UpdateButton() end
          SetTooltipMoney( GameTooltip, profit )
     else
          GameTooltip:AddLine( L["You have no junk items in your inventory."], 1.0, 1.0, 1.0, 1 )
     end
     GameTooltip:Show()
end

-- AutoProfitX:OnLeaveButton()
-- =========================================================
function AutoProfitX:OnLeaveButton()
     GameTooltip:Hide()
     if ( AutoProfitX.dbpc.char.buttonSpin ~= 3 ) then AutoProfitX:UpdateButton( "stop" ) end
end

-- AutoProfitX:OnClickButton()
-- =========================================================
function AutoProfitX:OnClickButton()
     local profit = AutoProfitX:GetProfit()
     if ( profit > 0 ) then
          GameTooltip:Hide()
          AutoProfitX:SellJunk()
          if ( AutoProfitX.dbpc.char.showTotal ) then
               if ( AutoProfitX.dbpc.char.fancyprofits ) then 
                    self:Print( "print", format( L["Total profits: %s"], AutoProfitX:GetMoneyString( profit ) ) );
               else
                    self:Print( "print", format( L["Total profits: %s"], AutoProfitX:CopperToGold( profit ) ) );
               end
          end
	  --- this is going to be removed SEE - AutoProfitX:ButtonState()
          if ( AutoProfitX.dbpc.char.buttonSpin ~= 2 ) then
               AutoProfitX:UpdateButton( "stop" )
          end
     end
end

-- AutoProfitX:OnShowButton()
-- =========================================================
function AutoProfitX:ButtonState()
     -- Needed for OnEnterButton
     if ( AutoProfitX.dbpc.char.buttonSpin == 1 ) then
          AutoProfitX:UpdateButton()
     end
     -- Needed for OnLeaveButton
     if ( AutoProfitX.dbpc.char.buttonSpin ~= 3 ) then
          AutoProfitX:UpdateButton( "stop" )
     end
     -- Needed for OnClick
     if ( AutoProfitX.dbpc.char.buttonSpin ~= 2 ) then
          AutoProfitX:UpdateButton( "stop" )
     end
     -- Needed for OnShowButton
     if ( AutoProfitX.dbpc.char.buttonSpin == 3 and AutoProfitX:GetProfit() > 0 ) then
          AutoProfitX:UpdateButton()
     end
end

-- AutoProfitX:OnShowButton()
-- =========================================================
function AutoProfitX:OnShowButton( self )
     if ( AutoProfitX.dbpc.char.autoSell ) then
          self:Hide()
     end
     if ( AutoProfitX.dbpc.char.buttonSpin == 3 and AutoProfitX:GetProfit() > 0 ) then
          AutoProfitX:UpdateButton()
     end
end

-- AutoProfitX:SetButtonXY( buttonXpos, buttonYpos )
-- =========================================================
function AutoProfitX:SetButtonXY( X, Y )
     X = tonumber( X )
     Y = tonumber( Y )
     if ( X ) then
          AutoProfitX.dbpc.char.buttonXpos = X
          self:Print( "bug", "X = "..X )
     end
     if ( Y ) then
          AutoProfitX.dbpc.char.buttonYpos = Y
          self:Print( "bug", "Y = "..Y )
     end
     AutoProfitX_SellButton:SetPoint( "TOPRIGHT", "MerchantFrame", "TOPRIGHT", AutoProfitX.dbpc.char.buttonXpos, AutoProfitX.dbpc.char.buttonYpos );
end

-- AutoProfitX:AddTooltipMoney( s )
-- =========================================================
function AutoProfitX:AddTooltipMoney( s )
     if ( s ) then
          totalProfit = totalProfit + s
     end
end

-- AutoProfitX:UpdateButton( action, spinRate )
-- =========================================================
function AutoProfitX:UpdateButton( action, spinRate )
     if ( action == "stop" ) then
          AutoProfitX_SellButton_TreasureModel.rotRate = 0;
     else
          spinRate = tonumber( spinRate )
          if ( not spinRate ) then
               spinRate = AutoProfitX.dbpc.char.spinRate
          end
          AutoProfitX_SellButton_TreasureModel.rotRate = spinRate;
     end

end

-- AutoProfitX:Print( action, msg )
-- =========================================================
function AutoProfitX:Print( action, msg )
     if ( action == "bug" ) then
          if ( AutoProfitX.db.global.debug.state ) then
               DEFAULT_CHAT_FRAME:AddMessage( format( L["|cff304aa9[AutoProfitX Debug]|r %s"], msg) );
          end
     elseif ( action == "print" ) then
              DEFAULT_CHAT_FRAME:AddMessage( format( L["|cff304aa9[AutoProfitX]|r %s"], msg ) )
     end
end

-- AutoProfitX:ButtonClick( action )
-- =========================================================
function AutoProfitX:ButtonClick( action )
     if ( action == "config" ) then
          InterfaceOptionsFrame_OpenToCategory(self.ConfigFrames.About);
     elseif ( action == "list" ) then
          if ( AutoProfitX.db.global.debug.state ) then
               ReloadUI();
          else
               self:Print( "print", L["You are using "..AddOnName.." by "]..GetAddOnMetadata("AutoProfitX", "Author") )
--               if ( not AutoProfitXFrame:IsVisible() ) then
--	            self:Print( "bug", "Show AutoProfitX Frame");
--                    --CompanionFinderFrame:Show();
--               else
--                    self:Print( "bug", "Hideing AutoProfitX Frame");
--                    --CompanionFinderFrame:Hide();
--               end
          end
     end
end

-- AutoProfitX:GetMoneyString( money )
-- =========================================================
function AutoProfitX:GetMoneyString( money )
     local gold = floor( money / 10000 )
     local silver = floor( ( money - gold * 10000 ) / 100 )
     local copper = mod( money, 100 )
     if ( gold > 0 ) then
          return format( GOLD_AMOUNT_TEXTURE.." "..SILVER_AMOUNT_TEXTURE.." "..COPPER_AMOUNT_TEXTURE, gold, 0, 0, silver, 0, 0, copper, 0, 0 )
     elseif ( silver > 0 ) then
          return format( SILVER_AMOUNT_TEXTURE.." "..COPPER_AMOUNT_TEXTURE, silver, 0, 0, copper, 0, 0 )
     else
          return format( COPPER_AMOUNT_TEXTURE, copper, 0, 0 )
     end
end

-- AutoProfitX:CopperToGold( copper, showGold )
-- =========================================================
function AutoProfitX:CopperToGold( copper )
     local strValue = ""
     local val
     val = math.floor( copper / COPPER_PER_GOLD )
     copper = mod( copper , COPPER_PER_GOLD )
     if ( val > 0 ) then
          strValue = val .. L["g "]
     end
     val = math.floor(copper/COPPER_PER_SILVER)
     copper = mod(copper,COPPER_PER_SILVER)
     if ( val > 0 or strValue ~= "" ) then
          strValue = strValue .. val .. L["s "]
     end
     return strValue .. copper .. L["c"]
end

-- AutoProfitX:OpenAllBags( Close )
-- =========================================================
function AutoProfitX:OpenAllBags( Close )
     for bag = 0, 4 do
          if ( not Close ) then
               if ( not IsBagOpen( bag ) ) then
                    AutoProfitX.db.global.bag[bag] = true;
                    OpenBag(bag);
               end
          elseif AutoProfitX.db.global.bag[bag] then
                    AutoProfitX.db.global.bag[bag] = false;
                    CloseBag(bag);
          end
     end
end

-- AutoProfitX:otherBag()
-- =========================================================
function AutoProfitX:otherBag()
     local otherBag = false;
     if ( IsAddOnLoaded("BaudBag") ) then self:Print( "bug", "BaudBag was found to be loaded"); otherBag = true; end
     if ( IsAddOnLoaded("ArkInventory") ) then self:Print( "bug", "ArkInventory was found to be loaded"); otherBag = true; end
     if ( IsAddOnLoaded("Bagnon") ) then self:Print( "bug", "Bagnon was found to be loaded"); otherBag = true; end
     if ( IsAddOnLoaded("OneBag3") ) then self:Print( "bug", "OneBag3 was found to be loaded"); otherBag = true; end
     return otherBag
end

-- AutoProfitX:tdeepcopy( from )
-- =========================================================
local function tdeepcopy( from )
     if ( type( from ) == "table" ) then
          local t = {}
          for k, v in pairs( from ) do
               if ( type(v) == "table" ) then
                    t[k] = tdeepcopy( v )
               else
                    t[k] = v
               end
          end
          return t
     end
     return from
end